/// This file is part of the X3DToolKit library
/// Copyright (C) 2002-2003  Yannick Le Goc (legoc@imag.fr)

/// http://www-imagis.imag.fr/Membres/Yannick.Legoc/X3D/


/// X3DToolKit is free software; you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation; either version 2 of the License, or
/// (at your option) any later version.

/// X3DToolKit is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.

/// You should have received a copy of the GNU General Public License
/// along with X3DToolKit; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
//                            X3DBaseTypes.h                                  //
////////////////////////////////////////////////////////////////////////////////

#ifndef X3DBASETYPES_H
#define X3DBASETYPES_H

#include "SFString.h"
#include "SFVec3f.h"
#include "SFPoint3f.h"
#include "SFMatrix34f.h"

#include <iostream>
#include <string>
#include <list>
#include <vector>
#include <map>

namespace X3DTK {

/// MFString class providing multiple SFString values. It derives from std::vector<SFString>.
class MFString : public std::vector<SFString>
{
public:
  /// Constructor.
  MFString(){};
  /// Constructor from a SFString.
  explicit MFString(const SFString &s);
};

/// Bool class
typedef bool SFBool;

/// MFBool class providing multiple SFBool values. It derives from std::vector<SFBool>.
class MFBool : public std::vector<SFBool>
{
public:
  /// Constructor.
  MFBool(){};
  /// Constructor from a SFString.
  explicit MFBool(const SFString &s);
};

/// SFInt32 class
typedef int SFInt32;

/// MFInt32 class providing multiple SFInt32 values. It derives from std::vector<SFInt32>.
class MFInt32 : public std::vector<SFInt32>
{
public:
  /// Constructor.
  MFInt32(){};
  /// Constructs a MFInt32 of size n.
  explicit MFInt32(size_type n, const SFInt32 &V = SFInt32()); 
  /// Constructor from a SFString.
  explicit MFInt32(const SFString &s);
};

/// Float class
typedef float SFFloat;

/// MFFloat class providing multiple SFFloat values. It derives from std::vector<SFFloat>.
class MFFloat : public std::vector<SFFloat>
{
public:
  /// Constructor.
  MFFloat(){};
  /// Constructs a MFFloat of size n.
  explicit MFFloat(size_type n, const SFFloat &V = SFFloat());
  /// Constructor from a SFString.
  explicit MFFloat(const SFString &s);
};

/// SFDouble class
typedef double SFDouble;

/// MFDouble class providing multiple SFDouble values. It derives from std::vector<SFDouble>.
class MFDouble : public std::vector<SFDouble>
{
public:
  /// Constructor.
  MFDouble(){};
  /// Constructor from a SFString.
  explicit MFDouble(const SFString &s);
};

/// SFColorRGBA class
class SFColorRGBA
{
public:
  /// Red canal.
  SFFloat r;
  /// Green canal.
  SFFloat g;
  /// Blue canal.
  SFFloat b;
  /// Alpha canal.
  SFFloat a;
  /// Constructor.  
  SFColorRGBA(){};
  /// Constructor from the attributes.  
  SFColorRGBA(SFFloat r, SFFloat g, SFFloat b, SFFloat a);
  /// Constructor from a SFString.
  explicit SFColorRGBA(const SFString &s);
  
  /// == operator.
  friend bool operator== (const SFColorRGBA &r1, const SFColorRGBA &r2);
  /// != operator.
  friend bool operator!= (const SFColorRGBA &r1, const SFColorRGBA &r2);
};

/// MFColorRGBA class providing multiple SFColorRGBA values. It derives from std::vector<SFColorRGBA>.
class MFColorRGBA : public std::vector<SFColorRGBA>
{
public:
  /// Constructor.
  MFColorRGBA(){};
  /// Constructs a MFColorRGBA of size n.
  MFColorRGBA(size_type n, const SFColorRGBA &C = SFColorRGBA());
  /// Constructor from a SFString.
  explicit MFColorRGBA(const SFString &s);
};

/// SFColor class
class SFColor
{
public:
  /// Red canal.
  SFFloat r;
  /// Green canal.
  SFFloat g;
  /// Blue canal.
  SFFloat b;
  /// Constructor.
  SFColor(){};
  /// Constructor from the attributes.
  SFColor(SFFloat r, SFFloat g, SFFloat b);
  /// Constructor from a SFString.
  explicit SFColor(const SFString &s);
  /// Converts the color to a SFColorRGBA.
  operator const SFColorRGBA() const;
 
  /// == operator.
  friend bool operator== (const SFColor &r1, const SFColor &r2);
  /// != operator.
  friend bool operator!= (const SFColor &r1, const SFColor &r2);
};


/// MFColor class providing multiple SFColor values. It derives from std::vector<SFColor>.
class MFColor : public std::vector<SFColor>
{
public:
  /// Constructor.  
  MFColor(){};
  /// Constructor from a SFString.
  explicit MFColor(const SFString &s);
  operator const MFColorRGBA() const;
};

class X3DAbstractNode;
/// SFAbstractNode class 
typedef X3DAbstractNode *SFAbstractNode;
/// MFAbstractNode.
typedef std::list<SFAbstractNode> MFAbstractNode;

namespace X3D {

class X3DNode;
/// SFNode class 
typedef X3DNode *SFNode;
/// MFNode.
typedef std::list<SFNode> MFNode;
/// SFNodeLink defines a couple of SFNode.
typedef std::pair<SFNode, SFNode> SFNodeLink;
/// MFDNode defines a list of SFNodeLink.
typedef std::list<SFNodeLink> MFDNode;

}

namespace GL {

class X3DNode;
/// SFNode class 
typedef X3DNode *SFNode;
/// MFNode.
typedef std::list<SFNode> MFNode;

}

/// SFVec2d class  
class SFVec2d
{
public:
  SFDouble x, y;
  /// Constructor.  
  SFVec2d();
  /// Constructor from the attributes.
  SFVec2d(SFDouble x, SFDouble y);
  /// Constructor from a SFString.
  explicit SFVec2d(const SFString &s);
};
  
/// MFVec2d class providing multiple SFVec2d values. It derives from std::vector<SFVec2d>.
class MFVec2d : public std::vector<SFVec2d>
{
public:
  /// Constructor.    
  MFVec2d(){};
  /// Constructor from a SFString.
  explicit MFVec2d(const SFString &s);
};
  
/// SFVec2f class 
class SFVec2f
{
public:
  SFFloat x, y;
  /// Constructor.    
  SFVec2f();
  /// Constructor from the attributes.    
  SFVec2f(SFFloat x, SFFloat y);
  /// Constructor from a SFString.
  explicit SFVec2f(const SFString &s);
};

/// MFVec2f class providing multiple SFVec2f values. It derives from std::vector<SFVec2f>.
class MFVec2f : public std::vector<SFVec2f>
{
public:
  /// Constructor.  
  MFVec2f(){};
  /// Constructor from a SFString.
  explicit MFVec2f(const SFString &s);
};

/// SFVec3d class
class SFVec3d
{
public:
  SFDouble x, y, z;
  /// Constructor.    
  SFVec3d();
  /// Constructor from the attributes.    
  SFVec3d(SFDouble x, SFDouble y, SFDouble z);
  /// Constructor from a SFString.
  explicit SFVec3d(const SFString &s);    
};

/// MFVec3d class providing multiple SFVec3d values. It derives from std::vector<SFVec3d>.
class MFVec3d : public std::vector<SFVec3d>
{
public:
  /// Constructor.  
  MFVec3d(){};
  /// Constructor from a SFString.
  explicit MFVec3d(const SFString &s);
};

/// MFVec3f class providing multiple SFVec3f values. It derives from std::vector<SFVec3f>. 
class MFVec3f : public std::vector<SFVec3f>
{
public:
  /// Constructor.  
  MFVec3f(){};
  /// Constructs a MFVec3f of size n.
  explicit MFVec3f(size_type n, const SFVec3f &V = SFVec3f());
  /// Constructor from a SFString.
  explicit MFVec3f(const SFString &s);
};

/// SFRotation 
class SFRotation
{
public:
  SFFloat x, y, z, angle;
  /// Constructor.  
  SFRotation(){};
  /// Constructor from the attributes.    
  SFRotation(SFFloat x, SFFloat y, SFFloat z, SFFloat angle);
  /// Constructor from a SFString.
  explicit SFRotation(const SFString &s);
  
  /// == operator.    
  friend bool operator== (const SFRotation &r1, const SFRotation &r2);
  /// != operator.
  friend bool operator!= (const SFRotation &r1, const SFRotation &r2);
};


/// MFRotation class providing multiple SFRotation values. It derives from std::vector<SFRotation>.
class MFRotation : public std::vector<SFRotation>
{
public:
  /// Constructor.
  MFRotation(){};
  /// Constructor from a SFString.
  explicit MFRotation(const SFString &s);
};


//constants
static const double PI = 3.14159265358979323846;
static const double radCoef = 180.0/PI;

/// function converting radians to degrees, used for GL rotations
inline SFFloat radiansToDegrees(SFFloat alpha)
{
  return radCoef*alpha;
}

#ifndef DOXYGEN

struct eqSFString
{
  inline bool operator()(const SFString &s1, const SFString &s2) const
  {
    return s1 == s2;
  }
};

struct ltSFString
{
  bool operator()(const SFString &s1, const SFString &s2) const
  {
    return strcmp(s1, s2) < 0;
  }
};

}

#endif

namespace X3DTK {

// other types
class Type;
class Component;
class CreationFunction;
class WalkingFunction;
class VisitingFunctions;
class EnterFunction;
class WalkOnFunction;
class LeaveFunction;

class X3DComponentCreator;
/// Pointer to CreationFunction.
typedef X3D::SFNode (X3DComponentCreator::* ptrToCreationFunction)() const;

class X3DComponentWalker;
/// Pointer to WalkingFunction.
typedef void (X3DComponentWalker::* ptrToWalkingFunction)(SFAbstractNode) const;

class X3DComponentVisitor;
/// Pointer to EnterFunction.
typedef void (X3DComponentVisitor::* ptrToEnterFunction)(SFAbstractNode) const;
/// Pointer to WalkOnFunction.
typedef bool (X3DComponentVisitor::* ptrToWalkOnFunction)(SFAbstractNode, SFAbstractNode) const;
/// Pointer to LeaveFunction.
typedef void (X3DComponentVisitor::* ptrToLeaveFunction)(SFAbstractNode) const;

/// MFType provides multiple values of Type.
typedef std::map<SFString, Type *, ltSFString> MFType;

/// MFComponent provides multiple values of Component.
typedef std::map<SFString, Component *, ltSFString> MFComponent;
/// Dict on CreationFunction.
typedef std::map<SFString, CreationFunction *, ltSFString> CreationDict;
/// Dict on WalkingFunction.
typedef std::map<SFString, WalkingFunction *, ltSFString> WalkingDict;
/// Dict on EnterFunction.
typedef std::map<SFString, EnterFunction *, ltSFString> EnterDict;
/// Dict on WalkOnFunction.
typedef std::map<SFString, WalkOnFunction *, ltSFString> WalkOnDict;
/// Dict on LeaveFunction.
typedef std::map<SFString, LeaveFunction *, ltSFString> LeaveDict;

/// Array of WalkingFunction.
typedef std::vector<WalkingFunction *> WalkingArray;
/// Array of VisitingFunction.
typedef std::vector<VisitingFunctions *> VisitingArray;

class X3DNodeProxy;
typedef std::list<X3DNodeProxy *> MFNodeProxy; 

/// operator<< overloaded for MFString class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::MFString &ms);
/// operator<< overloaded for MFBool class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::MFBool &mb);
/// operator<< overloaded for MFint32 class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::MFInt32 &mi);
/// operator<< overloaded for MFFloat class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::MFFloat &mf);
/// operator<< overloaded for MFDouble class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::MFDouble &md);
/// operator<< overloaded for SFColor class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::SFColor &c);
/// operator<< overloaded for MFColor class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::MFColor &mc);
/// operator<< overloaded for SFColorRGBA class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::SFColorRGBA &c);
/// operator<< overloaded for MFColorRGBA class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::MFColorRGBA &mc);
/// operator<< overloaded for SFVec2d class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::SFVec2d &v);
/// operator<< overloaded for MFVec2d class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::MFVec2d &mv);
/// operator<< overloaded for SFVec2f class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::SFVec2f &v);
/// operator<< overloaded for MFVec2f class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::MFVec2f &mv);
/// operator<< overloaded for SFVec3d class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::SFVec3d &v);
/// operator<< overloaded for MFVec3d class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::MFVec3d &mv);
/// operator<< overloaded for SFVec3f class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::SFVec3f &v);
/// operator<< overloaded for MFVec3f class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::MFVec3f &mv);
/// operator<< overloaded for SFPoint3f class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::SFPoint3f &v);
/// operator<< overloaded for SFRotation class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::SFRotation &r);
/// operator<< overloaded for MFRotation class respecting the X3D notation
std::ostream& operator<<(std::ostream& o, const X3DTK::MFRotation &mr);

}

#endif
